/************************************************************************
 * Program:        GLgalaxy
 * File:           GLgalaxy.c
 * Date Created:   2/01
 * Orig Author:    Uli Siegmund for Amiga
 * OpenGL Author:  Jeremy Witt wittj@onramp.net
 * Desc:           This is an implementation of the Classic Galaxy 
 *                 Screensaver which has been ported to a 3d Environment
 *                 using OpenGL
 *
 * Historical Information:
 *
 * Originally done by Uli Siegmund <uli@wombat.okapi.sub.org> on Amiga
 *   for EGS in Cluster
 * Port from Cluster/EGS to C/Intuition by Harald Backert
 * Port to X11 and incorporation into xlockmore by Hubert Feyrer
 *   <hubert.feyrer@rz.uni-regensburg.de>
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Revision History:
 * 15-Feb-01: wittj@onramp.net Ported to OpenGL + various enhancements
 * 26-Aug-00: robert.nagtegaal@phil.uu.nl and roland@tschai.demon.nl:
 *            various improvements
 * 10-May-97: jwz@jwz.org: turned into a standalone program.
 * 18-Apr-97: Memory leak fixed by Tom Schmidt <tschmidt@micron.com>
 * 07-Apr-97: Modified by Dave Mitchell <davem@magnet.com>
 * 23-Oct-94: Modified by David Bagley <bagleyd@bigfoot.com>
 *  random star sizes
 *  colors change depending on velocity
 * 10-Oct-94: Add colors by Hubert Feyer
 * 30-Sep-94: Initial port by Hubert Feyer
 * 09-Mar-94: VMS can generate a random number 0.0 which results in a
 *            division by zero, corrected by Jouk Jansen
 *            <joukj@crys.chem.uva.nl>
 *
 ***********************************************************************/
 
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <GL/glx.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <time.h>
#include <X11/extensions/xf86vmode.h>
#include <X11/keysym.h>

/*************************************************
 Defines and Macros Needed for GALAXY Algorithm
 -Modified somewhat for my use- JW
 *************************************************/
#define LRAND()			((long) (rand() & 0x7fffffff))
#define NRAND(n)		((int) (LRAND() % (n)))
#define MAXRAND			(2147483648.0) /* unsigned 1<<31 as a float */
#define SRAND(n)		/* already seeded by screenhack.c */

#define SINF(n)			((float)sin((double)(n)))
#define COSF(n)			((float)cos((double)(n)))
#define FABSF(n)		((float)fabs((double)(n)))

#undef MAX
#undef MIN
#undef ABS
#define MAX(a,b)((a)>(b)?(a):(b))
#define MIN(a,b)((a)<(b)?(a):(b))
#define ABS(a)((a)<0 ? -(a) : (a))

#define FLOATRAND ((double) LRAND() / ((double) MAXRAND))
#define M_PI 3.1415926535
#define MINSIZE       1
#define MINGALAXIES    2
#define MAX_STARS    1500
#define MAX_IDELTAT    50
#define EPSILON 0.00000001
#define sqrt_EPSILON 0.0001
#define DELTAT (MAX_IDELTAT * 0.0001)
#define GALAXYRANGESIZE  0.1
#define GALAXYMINSIZE  0.15
#define QCONS    0.001
#define COLORBASE  256

#define RADTODEG 180/M_PI

/***********************************************************
 Structures needed by Galaxy Algorithm
 ***********************************************************/
typedef struct 
{
     double      pos[3], vel[3];
} Star;

typedef struct 
{
     int         mass;
     int         nstars;
     Star       *stars;
     double  pos[3], vel[3];
     int     galcolr;
     int     galcolb;
     int     galcolg;
} Galaxy;

typedef struct 
{
 double      mat[3][3]; /* Movement of stars(?) */
 double      scale; /* Scale */
 double      size; /* */
 double      diff[3]; /* */
 Galaxy     *galaxies; /* the Whole Universe */
 int         ngalaxies; /* # galaxies */
 int         f_hititerations; /* # iterations before restart */
 int         step;
 double      rot_y; /* rot of eye around center of universe, around y-axis*/
 double      rot_x; /* rot of eye around center of universe, around x-axis */
} unistruct;

static unistruct *universes = NULL;

/************************************************************ 
 stuff about our window grouped together 
 ************************************************************/
typedef struct {
    Display *dpy;
    int screen;
    Window win;
    GLXContext ctx;
    XSetWindowAttributes attr;
    Bool fs;
    Bool doubleBuffered;
    XF86VidModeModeInfo deskMode;
    int x, y;
    unsigned int width, height;
    unsigned int depth;    
} GLWindow;

/*Holds our Texture between loading from BMP and loading to GL*/
typedef struct
{
     int width;
     int height;
     unsigned char *data;
} textureImage;

/* attributes for a single buffered visual in RGBA format with at least
 * 4 bits per color and a 16 bit depth buffer */
static int attrListSgl[] = {GLX_RGBA, GLX_RED_SIZE, 4, 
    GLX_GREEN_SIZE, 4, 
    GLX_BLUE_SIZE, 4, 
    GLX_DEPTH_SIZE, 16,
    None};

/* attributes for a double buffered visual in RGBA format with at least
 * 4 bits per color and a 16 bit depth buffer */
static int attrListDbl[] = { GLX_RGBA, GLX_DOUBLEBUFFER, 
    GLX_RED_SIZE, 4, 
    GLX_GREEN_SIZE, 4, 
    GLX_BLUE_SIZE, 4, 
    GLX_DEPTH_SIZE, 16,
    None };

GLWindow GLWin;

/***********************************************************
 Globals
 ***********************************************************/
Bool    done;                   /* used to designate the time to stop */
GLfloat zoom;         /* Zoom use PG-UP/PG-Down */

GLuint  texture[2];             /* Texture pointers */
GLuint  starlist;               /* GLList for stars */
Bool    twinkle;                /* not used -yet*/
GLfloat tw;               /* Size of stars*/

int     frame=0;                /* used for FPS calculation */
int     ourtime;                /* used for FPS calcualtion */
int     timebase=0;             /* used for FPS calculation */
Bool    fpstoggle=False;        /* default is FPS on */
GLuint  base;                   /* base display list of our font set */

GLuint  rendermode=1;           /* Classical (0) or 3D (1) */


/***********************************************************
 buildFont 
 -Loads a Font and builds a GLList for display
 ***********************************************************/
GLvoid buildFont(GLvoid)
{
    XFontStruct *font;

    base = glGenLists(96);      /* storage for 96 characters */
    /* load a font with a specific name  "Host Portable Character Encoding" */
    font = XLoadQueryFont(GLWin.dpy,
        "-*-helvetica-bold-r-normal--10-*-*-*-p-*-iso8859-1");
    if (font == NULL)
    {
        /* this really *should* be available on every X Window System...*/
        font = XLoadQueryFont(GLWin.dpy, "fixed");
        if (font == NULL)
        {
            printf("Problems loading fonts :-(\n");
            exit(1);
        }
    }
    /* build 96 display lists out of our font starting at char 32 */
    glXUseXFont(font->fid, 32, 96, base);
    /* free our XFontStruct since we have our display lists */
    XFreeFont(GLWin.dpy, font);
}

/***********************************************************
 deleteFont 
 -Deletes used GLList
 ***********************************************************/
GLvoid deleteFont(GLvoid)
{
    glDeleteLists(base, 96);
}

/***********************************************************
 printGLf
 -acts like printf and prints formatted txt to screen using 
  font 'base'
 ***********************************************************/
GLvoid printGLf(const char *fmt, ...)
{
    va_list ap;     /* our argument pointer */
    char text[256];
    if (fmt == NULL)    /* if there is no string to draw do nothing */
    {   
         printf("Some error here\n");
         return;
    }
    va_start(ap, fmt);  /* make ap point to first unnamed arg */
    vsprintf(text, fmt, ap);
    va_end(ap);
    glPushAttrib(GL_LIST_BIT);
    glListBase(base - 32);
    glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);
    glPopAttrib();
}

/***********************************************************
 free_galaxies
 -free memory used by galaxy structure
 ***********************************************************/
static void
free_galaxies(unistruct * gp)
{
 if (gp->galaxies != NULL) {
  int         i;

  for (i = 0; i < gp->ngalaxies; i++) 
  {
       Galaxy     *gt = &gp->galaxies[i];
       
       if (gt->stars != NULL)
            (void) free((void *) gt->stars);
  }
  (void) free((void *) gp->galaxies);
  gp->galaxies = NULL;
 }
}

/***********************************************************
 startover
 -reinitializes the universe
 ***********************************************************/
static void
startover()
{
     unistruct  *gp = universes;
     int         i, j;          /* tmp */
     double      w1, w2;        /* more tmp */
     double      d, v, w, h;    /* yet more tmp */

     gp->step = 0;
     gp->rot_y = 0;
     gp->rot_x = 0;
    
     gp->ngalaxies = NRAND(5);  /* Random galaxies from MIN to 5 */
     if (gp->ngalaxies < -MINGALAXIES)
          gp->ngalaxies = NRAND(-gp->ngalaxies - MINGALAXIES + 1) 
               + MINGALAXIES;
     else if (gp->ngalaxies < MINGALAXIES)
          gp->ngalaxies = MINGALAXIES;
     if (gp->galaxies == NULL)
          gp->galaxies = (Galaxy *) calloc(gp->ngalaxies, sizeof (Galaxy));
 
     for (i = 0; i < gp->ngalaxies; ++i) 
     {
          Galaxy     *gt = &gp->galaxies[i];
          double      sinw1, sinw2, cosw1, cosw2;

          gt->galcolr= NRAND(COLORBASE - 2);
          gt->galcolb= NRAND(COLORBASE - 2);
          gt->galcolg= NRAND(COLORBASE - 2);

          if (gt->stars != NULL) 
          {
               (void) free((void *) gt->stars);
               gt->stars = NULL;
          }
          gt->nstars = (NRAND(MAX_STARS / 2)) + MAX_STARS / 2;
          gt->stars = (Star *) malloc(gt->nstars * sizeof (Star));

          w1 = 2.0 * M_PI * FLOATRAND;
          w2 = 2.0 * M_PI * FLOATRAND;
          sinw1 = SINF(w1);
          sinw2 = SINF(w2);
          cosw1 = COSF(w1);
          cosw2 = COSF(w2);

          gp->mat[0][0] = cosw2;
          gp->mat[0][1] = -sinw1 * sinw2;
          gp->mat[0][2] = cosw1 * sinw2;
          gp->mat[1][0] = 0.0;
          gp->mat[1][1] = cosw1;
          gp->mat[1][2] = sinw1;
          gp->mat[2][0] = -sinw2;
          gp->mat[2][1] = -sinw1 * cosw2;
          gp->mat[2][2] = cosw1 * cosw2;

          gt->vel[0] = FLOATRAND * 2.0 - 1.0;
          gt->vel[1] = FLOATRAND * 2.0 - 1.0;
          gt->vel[2] = FLOATRAND * 2.0 - 1.0;
          gt->pos[0] = -gt->vel[0] * DELTAT * gp->f_hititerations 
               + FLOATRAND -0.5;
          gt->pos[1] = -gt->vel[1] * DELTAT * gp->f_hititerations 
               + FLOATRAND -0.5;
          gt->pos[2] = -gt->vel[2] * DELTAT * gp->f_hititerations 
               + FLOATRAND -0.5;

          gt->mass = (int) (FLOATRAND * 1000.0) + 1;
          
          gp->size = GALAXYRANGESIZE * FLOATRAND + GALAXYMINSIZE;
          
          for (j = 0; j < gt->nstars; ++j) 
          {
               Star       *st = &gt->stars[j];

               double      sinw, cosw;

               w = 2.0 * M_PI * FLOATRAND;
               sinw = SINF(w);
               cosw = COSF(w);
               d = FLOATRAND * gp->size;
               h = FLOATRAND * exp(-2.0 * (d / gp->size)) / 5.0 * gp->size;
               if (FLOATRAND < 0.5)
                    h = -h;
               st->pos[0] = gp->mat[0][0] * d * cosw + 
                            gp->mat[1][0] * d * sinw +
                            gp->mat[2][0] * h + gt->pos[0];

               st->pos[1] = gp->mat[0][1] * d * cosw + 
                            gp->mat[1][1] * d * sinw +
                            gp->mat[2][1] * h + gt->pos[1];

               st->pos[2] = gp->mat[0][2] * d * cosw + 
                            gp->mat[1][2] * d * sinw +
                            gp->mat[2][2] * h + gt->pos[2];
           
               v = sqrt(gt->mass * QCONS / sqrt(d * d + h * h));

               st->vel[0] = -gp->mat[0][0] * v * sinw + 
                             gp->mat[1][0] * v * cosw +
                             gt->vel[0];

               st->vel[1] = -gp->mat[0][1] * v * sinw + 
                             gp->mat[1][1] * v * cosw +
                             gt->vel[1];

               st->vel[2] = -gp->mat[0][2] * v * sinw + 
                             gp->mat[1][2] * v * cosw +
                             gt->vel[2];
           
               st->vel[0] *= DELTAT;
               st->vel[1] *= DELTAT;
               st->vel[2] *= DELTAT;
          }     
     }
}

/***********************************************************
 init_galaxy
 -called once upon program init (during initGL)
 ***********************************************************/
void
init_galaxy()
{
     unistruct  *gp;

     if (universes == NULL) 
     {
          if ((universes = (unistruct *) malloc(sizeof (unistruct))) == NULL)
               return;
     }
     gp = universes;

     gp->f_hititerations = 200; /* *JW* Make adjustable */
     gp->galaxies = NULL; /* *JW* Added because malloc doesn't return 0 filled mem! */

     switch(rendermode)
     {
     case 0:
          gp->scale = (double) (50.0+50.0)/8.0; /*JW*/
          break;
     case 1:
          gp->scale = (double) 300.0/8.0;
          break;
     }
     startover();
}

/***********************************************************
 draw_galaxy
 -Called for every frame, updates and draws Galaxy data
  called from drawGL
 ***********************************************************/
void
draw_galaxy()
{
     unistruct  *gp = universes;
     double      d, eps, cox=0, six=0, cor=0, sir=0;  /* tmp */
     int         i, j, k; /* more tmp */

     GLfloat     xproj,yproj;   /* used for X,Y projection of XYZ stuff */
     
     gp->rot_y += 0.01;
     gp->rot_x += 0.004;

     /*only calculate projection rotation stuff if we're doing
       classical mode*/
     if (rendermode == 0)
     {
          cox = COSF(gp->rot_y);
          six = SINF(gp->rot_y);
          cor = COSF(gp->rot_x);
          sir = SINF(gp->rot_x);
     }

     eps = 1/(EPSILON * sqrt_EPSILON * DELTAT * DELTAT * QCONS);

     for (i = 0; i < gp->ngalaxies; ++i) 
     {
          Galaxy     *gt = &gp->galaxies[i];
          glColor4ub(gt->galcolr,gt->galcolg,gt->galcolb,255);

          for (j = 0; j < gp->galaxies[i].nstars; ++j) 
          {
               Star       *st = &gt->stars[j];
               double      v0 = st->vel[0];
               double      v1 = st->vel[1];
               double      v2 = st->vel[2];

               for (k = 0; k < gp->ngalaxies; ++k) 
               {
                    Galaxy     *gtk = &gp->galaxies[k];
                    double      d0 = gtk->pos[0] - st->pos[0];
                    double      d1 = gtk->pos[1] - st->pos[1];
                    double      d2 = gtk->pos[2] - st->pos[2];

                    d = d0 * d0 + d1 * d1 + d2 * d2;
                    if (d > EPSILON)
                         d = gt->mass / (d * sqrt(d)) * DELTAT * DELTAT * QCONS;
                    else
                         d = gt->mass * eps;
                    v0 += d0 * d;
                    v1 += d1 * d;
                    v2 += d2 * d;
               }

               st->vel[0] = v0;
               st->vel[1] = v1;
               st->vel[2] = v2;
               
               st->pos[0] += v0;
               st->pos[1] += v1;
               st->pos[2] += v2;

               glLoadIdentity();
               glTranslatef(0.0f,0.0f,zoom);
               switch (rendermode)
               {
               case 0:
                   /*Make the projection onto a 2d surface 
                     (it just looks 3d)
                     This was converted from screen coord 
                     version to this one*/

                    xproj = (GLfloat) (((cox * st->pos[0]) - 
                                        (six * st->pos[2])) * gp->scale);
                    yproj = (GLfloat) (((cor * st->pos[1]) - 
                                        (sir * ((six * st->pos[0]) +
                                         (cox * st->pos[2])))) * gp->scale);
                    glTranslatef(xproj,yproj,0.0f); /* move to sprite loc */
                    break;
               case 1:
                    /* rotate universe */
                    glRotatef(gp->rot_y*RADTODEG,0.0f,1.0f,0.0f); 
                    /* Translate to Sprite Location */
                    glTranslatef(st->pos[0],st->pos[1],st->pos[2]);
                    /* Cancel rotation of Sprite (so it faces user)*/
                    glRotatef(-gp->rot_y*RADTODEG,0.0f,1.0f,0.0f);
                    break;
               }
               glCallList(starlist); /* draw sprite */
          }
          for (k = i + 1; k < gp->ngalaxies; ++k) 
          {
               Galaxy     *gtk = &gp->galaxies[k];
               double      d0 = gtk->pos[0] - gt->pos[0];
               double      d1 = gtk->pos[1] - gt->pos[1];
               double      d2 = gtk->pos[2] - gt->pos[2];

               d = d0 * d0 + d1 * d1 + d2 * d2;
               if (d > EPSILON)
                    d = gt->mass * gt->mass / (d * sqrt(d)) * DELTAT * QCONS;
               else
                    d = gt->mass*gt->mass/(EPSILON*sqrt_EPSILON)*DELTAT*QCONS;

               d0 *= d;
               d1 *= d;
               d2 *= d;
               gt->vel[0] += d0 / gt->mass;
               gt->vel[1] += d1 / gt->mass;
               gt->vel[2] += d2 / gt->mass;
               gtk->vel[0] -= d0 / gtk->mass;
               gtk->vel[1] -= d1 / gtk->mass;
               gtk->vel[2] -= d2 / gtk->mass;
          }
          
          gt->pos[0] += gt->vel[0] * DELTAT;
          gt->pos[1] += gt->vel[1] * DELTAT;
          gt->pos[2] += gt->vel[2] * DELTAT;
  
     }

     gp->step++;
     if (gp->step > gp->f_hititerations * 4)
     {
          free_galaxies(gp);
          startover();
     }
}

/***********************************************************
 release_galaxy
 -called once during program exit, cleans up memory
 ***********************************************************/
void
release_galaxy()
{
 if (universes != NULL) {
   free_galaxies(universes);
  (void) free((void *) universes);
  universes = NULL;
 }
}

/***********************************************************
 LoadBMP 
 -simple loader for 24bit bitmaps (data is in rgb-format
 ***********************************************************/
int LoadBMP(char *filename, textureImage *texture)
{
    FILE *file;
    unsigned short int bfType;
    long int bfOffBits;
    short int biPlanes;
    short int biBitCount;
    long int biSizeImage;
    int i;
    unsigned char temp;
    /* make sure the file is there and open it read-only (binary) */
    if ((file = fopen(filename, "rb")) == NULL)
    {
        printf("File not found : %s\n", filename);
        return 0;
    }
    if(!fread(&bfType, sizeof(short int), 1, file))
    {
        printf("Error reading file!\n");
        return 0;
    }
    /* check if file is a bitmap */
    if (bfType != 19778)
    {
        printf("Not a Bitmap-File!\n");
        return 0;
    }        
    /* get the file size */
    /* skip file size and reserved fields of bitmap file header */
    fseek(file, 8, SEEK_CUR);
    /* get the position of the actual bitmap data */
    if (!fread(&bfOffBits, sizeof(long int), 1, file))
    {
        printf("Error reading file!\n");
        return 0;
    }
    printf("Data at Offset: %ld\n", bfOffBits);
    /* skip size of bitmap info header */
    fseek(file, 4, SEEK_CUR);
    /* get the width of the bitmap */
    fread(&texture->width, sizeof(int), 1, file);
    printf("Width of Bitmap: %d\n", texture->width);
    /* get the height of the bitmap */
    fread(&texture->height, sizeof(int), 1, file);
    printf("Height of Bitmap: %d\n", texture->height);
    /* get the number of planes (must be set to 1) */
    fread(&biPlanes, sizeof(short int), 1, file);
    if (biPlanes != 1)
    {
        printf("Error: number of Planes not 1!\n");
        return 0;
    }
    /* get the number of bits per pixel */
    if (!fread(&biBitCount, sizeof(short int), 1, file))
    {
        printf("Error reading file!\n");
        return 0;
    }
    printf("Bits per Pixel: %d\n", biBitCount);
    if (biBitCount != 24)
    {
        printf("Bits per Pixel not 24\n");
        return 0;
    }
    /* calculate the size of the image in bytes */
    biSizeImage = texture->width * texture->height * 3;
    printf("Size of the image data: %ld\n", biSizeImage);
    texture->data = malloc(biSizeImage);
    /* seek to the actual data */
    fseek(file, bfOffBits, SEEK_SET);
    if (!fread(texture->data, biSizeImage, 1, file))
    {
        printf("Error loading file!\n");
        return 0;
    }
    /* swap red and blue (bgr -> rgb) */
    for (i = 0; i < biSizeImage; i += 3)
    {
        temp = texture->data[i];
        texture->data[i] = texture->data[i + 2];
        texture->data[i + 2] = temp;
    }
    return 1;
}

/***********************************************************
 LoadGLTextures
 -calls LoadBMP and converts to GL textures and loads them
 ***********************************************************/
Bool LoadGLTextures()   /* Load Bitmaps And Convert To Textures */
{
    Bool status;
    textureImage *texti;
    
    status = False;
    texti = malloc(sizeof(textureImage));
    if (LoadBMP("Data/Star.bmp", texti))
    {
        status = True;
        glGenTextures(1, &texture[0]);   /* create the texture */
        glBindTexture(GL_TEXTURE_2D, texture[0]);
        /* actually generate the texture */
        glTexImage2D(GL_TEXTURE_2D, 0, 3, texti->width, texti->height, 0,
            GL_RGB, GL_UNSIGNED_BYTE, texti->data);
        /* use linear filtering */
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    if (LoadBMP("Data/white.bmp", texti))
    {
        status = True;
        glGenTextures(1, &texture[1]);   /* create the texture */
        glBindTexture(GL_TEXTURE_2D, texture[1]);
        /* actually generate the texture */
        glTexImage2D(GL_TEXTURE_2D, 0, 3, texti->width, texti->height, 0,
            GL_RGB, GL_UNSIGNED_BYTE, texti->data);
        /* use linear filtering */
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    /* free the ram we used in our texture generation process */
    if (texti)
    {
        if (texti->data)
            free(texti->data);
        free(texti);
    }    
    return status;
}


/***********************************************************
 resizeGLScene 
 -called when window is resized
 ***********************************************************/
void resizeGLScene(unsigned int width, unsigned int height)
{
    if (height == 0)    /* Prevent A Divide By Zero If The Window Is Too Small */
        height = 1;
    glViewport(0, 0, width, height);    /* Reset The Current Viewport And Perspective Transformation */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, (GLfloat)width / (GLfloat)height, 0.1f, 100.0f);
    glMatrixMode(GL_MODELVIEW);
}

void init_vars(int rmode)
{
     switch (rmode)
     {
     case 0:
          zoom = -100.0f;
          tw   =  0.35f;
          break;
     case 1:
          zoom = -6.0f;
          tw   = 0.03f;
          break;
     }
}

/***********************************************************
 initGL
 -GL initialization function calls init_galaxy
 ***********************************************************/
int initGL(GLvoid)
{
    if (!LoadGLTextures())
    {
         return False;
    }
    /*setup our Global Vars based on the mode*/
    init_vars(rendermode);

    /*JW Generate BuildList for a star*/
    srand((unsigned int) time(NULL));
    starlist=glGenLists(1);
    glNewList(starlist,GL_COMPILE);    
      glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(1.0f, 1.0f); glVertex3f( tw, tw, 0.0f);
        glTexCoord2f(0.0f, 1.0f); glVertex3f(-tw, tw, 0.0f);
        glTexCoord2f(1.0f, 0.0f); glVertex3f( tw,-tw, 0.0f);
        glTexCoord2f(0.0f, 0.0f); glVertex3f(-tw,-tw, 0.0f);
      glEnd();
    glEndList();

    /*Setup other junk*/
    glEnable(GL_TEXTURE_2D);
    glShadeModel(GL_SMOOTH);
    glClearColor(0.0f, 0.0f, 0.0f, 0.5f);
    glClearDepth(1.0f);
    
    /*glEnable(GL_DEPTH_TEST);
      glDepthFunc(GL_LEQUAL); */ /*JWitt disable for blending */
    
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    glEnable(GL_BLEND);


    init_galaxy(); /*JW NEW*/
  

    buildFont();

    /* we use resizeGLScene once to set up our initial perspective */
    resizeGLScene(GLWin.width, GLWin.height);
    glFlush();
    return True;
}


/***********************************************************
 drawGLScene
 -Called once per frame, calls draw_galaxy, draws FPS if active
 ***********************************************************/
int drawGLScene(GLvoid)
{
    char s[80];

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   
    glBindTexture(GL_TEXTURE_2D, texture[0]);
    draw_galaxy();

    /*FPS*/
    if (fpstoggle)
    {
         frame++;
         ourtime=glutGet(GLUT_ELAPSED_TIME);
         if (ourtime - timebase > 1000)
         {
              sprintf(s,"avgFPS=%4.2f",
                      frame*1000.0/(ourtime-timebase));
              timebase = ourtime;		
              frame = 0;
              /*  printf("%s\n",s);*/
         }
         glBindTexture(GL_TEXTURE_2D, texture[1]);
         glColor4f(0.5f,0.0f,0.0f,0.5f);
         glLoadIdentity();
         glTranslatef(-5.5f,4.0f,-10.0f);
         glRasterPos2f(0.0f,0.0f);
         printGLf("GLgalaxy: %s",s);
    } /*END OF FPS*/
    
    if (GLWin.doubleBuffered)
    {
        glXSwapBuffers(GLWin.dpy, GLWin.win);
    }
    return True;    
}


/***********************************************************
 killGLWindow
 -Called once per frame, calls draw_galaxy, draws FPS if active
 ***********************************************************/
GLvoid killGLWindow(GLvoid)
{
    if (GLWin.ctx)
    {
        if (!glXMakeCurrent(GLWin.dpy, None, NULL))
        {
            printf("Could not release drawing context.\n");
        }
        glXDestroyContext(GLWin.dpy, GLWin.ctx);
        GLWin.ctx = NULL;
    }
    /* switch back to original desktop resolution if we were in fs */
    if (GLWin.fs)
    {
        XF86VidModeSwitchToMode(GLWin.dpy, GLWin.screen, &GLWin.deskMode);
        XF86VidModeSetViewPort(GLWin.dpy, GLWin.screen, 0, 0);
    }
    deleteFont();
    XCloseDisplay(GLWin.dpy);
}


/***********************************************************
 createGLWindow
 -setup a GL Window FIXME: bits is currently unused
 ***********************************************************/
Bool createGLWindow(char* title, int width, int height, int bits,
                    Bool fullscreenflag)
{
    XVisualInfo *vi;
    Colormap cmap;
    int dpyWidth, dpyHeight;
    int i;
    int glxMajorVersion, glxMinorVersion;
    int vidModeMajorVersion, vidModeMinorVersion;
    XF86VidModeModeInfo **modes;
    int modeNum;
    int bestMode;
    Atom wmDelete;
    Window winDummy;
    unsigned int borderDummy;
    
    GLWin.fs = fullscreenflag;
    /* set best mode to current */
    bestMode = 0;
    /* get a connection */
    GLWin.dpy = XOpenDisplay(0);
    GLWin.screen = DefaultScreen(GLWin.dpy);
    XF86VidModeQueryVersion(GLWin.dpy, &vidModeMajorVersion,
        &vidModeMinorVersion);
    printf("XF86VidModeExtension-Version %d.%d\n", vidModeMajorVersion,
        vidModeMinorVersion);
    XF86VidModeGetAllModeLines(GLWin.dpy, GLWin.screen, &modeNum, &modes);
    /* save desktop-resolution before switching modes */
    GLWin.deskMode = *modes[0];
    /* look for mode with requested resolution */
    for (i = 0; i < modeNum; i++)
    {
        if ((modes[i]->hdisplay == width) && (modes[i]->vdisplay == height))
        {
            bestMode = i;
        }
    }
    /* get an appropriate visual */
    vi = glXChooseVisual(GLWin.dpy, GLWin.screen, attrListDbl);
    if (vi == NULL)
    {
        vi = glXChooseVisual(GLWin.dpy, GLWin.screen, attrListSgl);
        GLWin.doubleBuffered = False;
        printf("Only Singlebuffered Visual!\n");
    }
    else
    {
        GLWin.doubleBuffered = True;
        printf("Got Doublebuffered Visual!\n");
    }
    glXQueryVersion(GLWin.dpy, &glxMajorVersion, &glxMinorVersion);
    printf("glX-Version %d.%d\n", glxMajorVersion, glxMinorVersion);
    /* create a GLX context */
    GLWin.ctx = glXCreateContext(GLWin.dpy, vi, 0, GL_TRUE);
    /* create a color map */
    cmap = XCreateColormap(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
        vi->visual, AllocNone);
    GLWin.attr.colormap = cmap;
    GLWin.attr.border_pixel = 0;

    if (GLWin.fs)
    {
        XF86VidModeSwitchToMode(GLWin.dpy, GLWin.screen, modes[bestMode]);
        XF86VidModeSetViewPort(GLWin.dpy, GLWin.screen, 0, 0);
        dpyWidth = modes[bestMode]->hdisplay;
        dpyHeight = modes[bestMode]->vdisplay;
        printf("Resolution %dx%d\n", dpyWidth, dpyHeight);
        XFree(modes);
    
        /* create a fullscreen window */
        GLWin.attr.override_redirect = True;
        GLWin.attr.event_mask = ExposureMask | KeyPressMask | ButtonPressMask |
            StructureNotifyMask;
        GLWin.win = XCreateWindow(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
            0, 0, dpyWidth, dpyHeight, 0, vi->depth, InputOutput, vi->visual,
            CWBorderPixel | CWColormap | CWEventMask | CWOverrideRedirect,
            &GLWin.attr);
        XWarpPointer(GLWin.dpy, None, GLWin.win, 0, 0, 0, 0, 0, 0);
        XMapRaised(GLWin.dpy, GLWin.win);
        XGrabKeyboard(GLWin.dpy, GLWin.win, True, GrabModeAsync,
            GrabModeAsync, CurrentTime);
        XGrabPointer(GLWin.dpy, GLWin.win, True, ButtonPressMask,
            GrabModeAsync, GrabModeAsync, GLWin.win, None, CurrentTime);
    }
    else
    {
        /* create a window in window mode*/
        GLWin.attr.event_mask = ExposureMask | KeyPressMask | ButtonPressMask |
            StructureNotifyMask;
        GLWin.win = XCreateWindow(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
            0, 0, width, height, 0, vi->depth, InputOutput, vi->visual,
            CWBorderPixel | CWColormap | CWEventMask, &GLWin.attr);
        /* only set window title and handle wm_delete_events if in windowed mode */
        wmDelete = XInternAtom(GLWin.dpy, "WM_DELETE_WINDOW", True);
        XSetWMProtocols(GLWin.dpy, GLWin.win, &wmDelete, 1);
        XSetStandardProperties(GLWin.dpy, GLWin.win, title,
            title, None, NULL, 0, NULL);
        XMapRaised(GLWin.dpy, GLWin.win);
    }       
    /* connect the glx-context to the window */
    glXMakeCurrent(GLWin.dpy, GLWin.win, GLWin.ctx);
    XGetGeometry(GLWin.dpy, GLWin.win, &winDummy, &GLWin.x, &GLWin.y,
        &GLWin.width, &GLWin.height, &borderDummy, &GLWin.depth);
    printf("Depth %d\n", GLWin.depth);
    if (glXIsDirect(GLWin.dpy, GLWin.ctx)) 
        printf("Congrats, you have Direct Rendering!\n");
    else
        printf("Sorry, no Direct Rendering possible!\n");
    initGL();
    return True;    
}

/***********************************************************
 keyPressed
 -key handler
 ***********************************************************/
void keyPressed(KeySym key)
{
     switch (key)
     {
     case XK_Escape:
          done=True;
          break;
     case XK_F1:
          killGLWindow();
          GLWin.fs = !GLWin.fs;
          createGLWindow("GLgalaxy -JWitt", 648, 480, 24, GLWin.fs);
          break;
     case XK_Page_Up:
          zoom -=0.2f;
/*          if (zoom < -100.0)
            zoom = -100.0;*/
          printf("Zoom= %f\n",zoom);
          break;
     case XK_Page_Down:
          zoom +=0.2f;
          printf("Zoom= %f\n",zoom);
          break;
     case XK_t:
          twinkle=!twinkle;
          break;
     case XK_Up:
          tw+=0.01f;
          glNewList(starlist,GL_COMPILE);    
          glBegin(GL_TRIANGLE_STRIP);
           glTexCoord2f(1.0f, 1.0f); glVertex3f( tw, tw, 0.0f);
           glTexCoord2f(0.0f, 1.0f); glVertex3f(-tw, tw, 0.0f);
           glTexCoord2f(1.0f, 0.0f); glVertex3f( tw,-tw, 0.0f);
           glTexCoord2f(0.0f, 0.0f); glVertex3f(-tw,-tw, 0.0f);
          glEnd();
          glEndList();
          printf("StarSize= %f\n",tw);
          break;
     case XK_Down:
          tw-=0.01f;
          glNewList(starlist,GL_COMPILE);    
          glBegin(GL_TRIANGLE_STRIP);
           glTexCoord2f(1.0f, 1.0f); glVertex3f( tw, tw, 0.0f);
           glTexCoord2f(0.0f, 1.0f); glVertex3f(-tw, tw, 0.0f);
           glTexCoord2f(1.0f, 0.0f); glVertex3f( tw,-tw, 0.0f);
           glTexCoord2f(0.0f, 0.0f); glVertex3f(-tw,-tw, 0.0f);
          glEnd();
          glEndList();
          printf("StarSize= %f\n",tw);
          break;
     case XK_f:
          fpstoggle=!fpstoggle;
          break;
     case XK_m:
          rendermode++;
          if (rendermode > 1)
               rendermode=0;
          init_vars(rendermode); /* set up vars for this mode */
          if (universes != NULL)
          {
               universes->scale= (double) 300.0/8.0;
               universes->step = universes->f_hititerations*4+1;
          }
          glNewList(starlist,GL_COMPILE);    
          glBegin(GL_TRIANGLE_STRIP);
           glTexCoord2f(1.0f, 1.0f); glVertex3f( tw, tw, 0.0f);
           glTexCoord2f(0.0f, 1.0f); glVertex3f(-tw, tw, 0.0f);
           glTexCoord2f(1.0f, 0.0f); glVertex3f( tw,-tw, 0.0f);
           glTexCoord2f(0.0f, 0.0f); glVertex3f(-tw,-tw, 0.0f);
          glEnd();
          glEndList();
          break;
 
     }
}

/***********************************************************
 main
 -this is the main function
 ***********************************************************/
int main(int argc, char **argv)
{
    XEvent event;
    
    done = False;
    /* default to fullscreen */
    GLWin.fs = True;
    createGLWindow("GLgalaxy - JWitt", 640, 480, 24, GLWin.fs);

    /* wait for events*/ 
    while (!done)
    {
        /* handle the events in the queue */
        while (XPending(GLWin.dpy) > 0)
        {
            XNextEvent(GLWin.dpy, &event);
            switch (event.type)
            {
                case Expose:
	                if (event.xexpose.count != 0)
	                    break;
                    drawGLScene();
         	        break;
            case ConfigureNotify:
            /* call resizeGLScene only if our window-size changed */
                if ((event.xconfigure.width != GLWin.width) || 
                    (event.xconfigure.height != GLWin.height))
                {
                    GLWin.width = event.xconfigure.width;
                    GLWin.height = event.xconfigure.height;
                        printf("Resize event\n");
                    resizeGLScene(event.xconfigure.width,
                        event.xconfigure.height);
                }
                break;
            /* exit in case of a mouse button press */
            case ButtonPress:     
                done = True;
                break;
            case KeyPress:
                 keyPressed(XLookupKeysym(&event.xkey,0));
                break;
            case ClientMessage:    
                if (*XGetAtomName(GLWin.dpy, event.xclient.message_type) == 
                    *"WM_PROTOCOLS")
                {
                    printf("Exiting sanely...\n");
                    done = True;
                }
                break;
            default:
                break;
            }
        }
        drawGLScene();
    }
    release_galaxy();
    killGLWindow();
    exit (0);
}
